import requests
import time
import logging
from cmp_version import cmp_version
try:
    from urllib.parse import urljoin
except ImportError:
    from urlparse import urljoin

_LOG = logging.getLogger("ubipop")


class UnsupportedTypeId(Exception):
    pass


class Pulp(object):
    PULP_API = "/pulp/api/v2/"

    def __init__(self, hostname, auth, insecure=False):
        self.hostname = hostname
        self.auth = auth
        self.scheme = "https://"
        self.base_url = urljoin(self.scheme + hostname, self.PULP_API)
        self.session = None
        self.insecure = insecure

    def do_request(self, req_type, url, data=None):
        if self.session is None:
            self.session = requests.Session()

            if len(self.auth) == 1:
                self.session.cert = self.auth[0]
            else:
                self.session.auth = self.auth

        req_url = urljoin(self.base_url, url)
        ret = None
        if req_type == "post":
            ret = self.session.post(req_url, json=data, verify=not self.insecure)
        elif req_type == "get":
            ret = self.session.get(req_url, verify=not self.insecure)

        return ret

    def search_repo_by_cs(self, content_set):
        url = "repositories/search/"
        payload = {"criteria": {"filters": {"notes.content_set": content_set}},
                   "distributors": True}

        ret = self.do_request("post", url, payload)
        ret.raise_for_status()
        repos = []
        for item in ret.json():
            notes = item['notes']
            repos.append(Repo(item['id'], notes['arch'], notes['platform_full_version'],
                              [(distributor['id'], distributor['distributor_type_id'])
                               for distributor in item['distributors']]))

        return repos

    def search_rpms(self, repo, name=None, arch=None, name_globbing=False):
        url = "repositories/{REPO_ID}/search/units/".format(REPO_ID=repo.repo_id)
        criteria = {"type_ids": ["rpm", "srpm"]}

        filters = {"filters": {"unit": {}}}
        if name:
            if name_globbing:
                filters["filters"]["unit"]["name"] = {"$regex": name + ".*"}
            else:
                filters["filters"]["unit"]["name"] = name

            criteria.update(filters)
        if arch:
            filters["filters"]["unit"]["arch"] = arch
            criteria.update(filters)

        payload = {"criteria": criteria}
        ret = self.do_request("post", url, payload)
        rpms = []
        ret.raise_for_status()
        for item in ret.json():
            metadata = item['metadata']
            rpms.append(Package(metadata['name'], metadata['filename'], metadata.get('sourcerpm'),
                                metadata.get('is_modular', False)))
        return rpms

    def search_modules(self, repo, name=None, stream=None):
        url = "repositories/{REPO_ID}/search/units/".format(REPO_ID=repo.repo_id)
        criteria = {"type_ids": ["modulemd"]}
        if name and stream:
            criteria.update({"filters": {"unit": {"name": name, "stream": stream}}})
        payload = {"criteria": criteria}

        ret = self.do_request("post", url, payload)
        modules = []
        ret.raise_for_status()
        for item in ret.json():

            metadata = item['metadata']
            modules.append(Module(metadata['name'], metadata['stream'],
                                  metadata['version'], metadata['context'],
                                  metadata['arch'], metadata['artifacts'], metadata['profiles']))
        return modules

    def search_module_defaults(self, repo, name=None, stream=None):
        url = "repositories/{REPO_ID}/search/units/".format(REPO_ID=repo.repo_id)
        criteria = {"type_ids": ["modulemd_defaults"]}
        if name and stream:
            criteria.update({"filters": {"unit": {"name": name, "stream": stream}}})
        payload = {"criteria": criteria}

        ret = self.do_request("post", url, payload)
        ret.raise_for_status()
        module_defaults = []
        for item in ret.json():
            metadata = item['metadata']
            module_defaults.append(ModuleDefaults(metadata['name'], metadata['stream'],
                                                  metadata['profiles']))
        return module_defaults

    def wait_for_tasks(self, task_id_list, delay=5.0):
        results = {}

        _tasks = set(task_id_list)
        while _tasks:
            statuses = self.search_tasks(_tasks)
            for status in statuses:
                if status["state"] in ("finished", "error", "cancelled"):
                    _tasks -= set([status["task_id"]])
                results[status["task_id"]] = status
            if _tasks:
                time.sleep(delay)
        return results

    def search_tasks(self, task_ids):
        url = "tasks/{task_id}/"
        statuses = []
        for task_id in task_ids:
            ret = self.do_request('get', url.format(task_id=task_id))
            statuses.append(ret.json())
        return statuses

    def _modules_query(self, modules):
        query_list = []
        for module in modules:
            query_list.append({'$and': [{'name': module.name},
                                        {'context': module.context},
                                        {'version': module.version},
                                        {'stream': module.stream},
                                        {'arch': module.arch}
                                        ]})

        return query_list

    def _module_defaults_query(self, module_defaults):
        query_list = []
        for md_d in module_defaults:
            query_list.append({'$and': [
                                    {'name': md_d.name},
                                    {'stream': md_d.stream},
                                    {'profiles': md_d.profiles}
                                ]})
        return query_list

    def _rpms_query(self, rpms):
        return [{"filename": rpm.filename} for rpm in rpms]

    def unassociate_units(self, repo, units, type_ids):
        url = "repositories/{dst_repo}/actions/unassociate/".format(dst_repo=repo.repo_id)
        data = {
            'criteria': {
                'type_ids': list(type_ids),
                'filters': {
                    'unit': {
                        "$or": self._get_query_list(type_ids, units)
                    }
                }
            },
        }
        log_msg = "Unassociating %s from %s"
        for unit in units:
            _LOG.info(log_msg, str(unit), repo.repo_id)

        ret = self.do_request('post', url, data).json()
        return [task['task_id'] for task in ret['spawned_tasks']]

    def associate_units(self, src_repo, dest_repo, units, type_ids):
        url = "repositories/{dst_repo}/actions/associate/".format(dst_repo=dest_repo.repo_id)
        data = {
            'source_repo_id': src_repo.repo_id,
            'criteria': {
                'type_ids': list(type_ids),
                'filters': {
                    'unit': {
                        '$or': self._get_query_list(type_ids, units),
                    },
                },
            },
        }
        log_msg = "Associating %s from %s to %s"
        for unit in units:
            _LOG.info(log_msg, str(unit), src_repo.repo_id, dest_repo.repo_id)
        ret = self.do_request('post', url, data)
        ret.raise_for_status()
        ret_json = ret.json()
        return [task['task_id'] for task in ret_json['spawned_tasks']]

    def _get_query_list(self, type_ids, units):
        if "modulemd" in type_ids:
            query_list = self._modules_query(units)
        elif "modulemd_defaults" in type_ids:
            query_list = self._module_defaults_query(units)
        elif "rpm" in type_ids or "srpm" in type_ids:
            query_list = self._rpms_query(units)
        else:
            raise UnsupportedTypeId

        return query_list

    def associate_modules(self, src_repo, dst_repo, modules):
        return self.associate_units(src_repo, dst_repo, modules, ("modulemd", ))

    def associate_module_defaults(self, src_repo, dst_repo, module_defaults):
        return self.associate_units(src_repo, dst_repo, module_defaults, ("modulemd_defaults", ))

    def associate_packages(self, src_repo, dst_repo, rpms):
        return self.associate_units(src_repo, dst_repo, rpms, ("rpm", "srpm"))

    def unassociate_modules(self, repo, modules):
        return self.unassociate_units(repo, modules, ("modulemd", ))

    def unassociate_module_defaults(self, repo, module_defaults):
        return self.unassociate_units(repo, module_defaults, ("modulemd_defaults", ))

    def unassociate_packages(self, repo, rpms):
        return self.unassociate_units(repo, rpms, ("rpm", "srpm"))

    def publish_repo(self, repo):
        url = "repositories/{repo_id}/actions/publish/".format(repo_id=repo.repo_id)
        task_ids = []
        for dist_id, dist_type_id in repo.distributors_ids_type_ids_tuples:
            _LOG.info("Publishing %s in %s", repo.repo_id, dist_id)
            data = {"id": dist_id}
            if dist_type_id in ("rpm_rsync_distributor", "cdn_distributor"):
                data["override_config"] = {"delete": True}
            ret = self.do_request('post', url, data).json()
            task_ids.extend(task['task_id'] for task in ret['spawned_tasks'])

        return task_ids


class Repo(object):
    def __init__(self, repo_id, arch, platform_full_version, distributors_ids_type_ids):
        self.repo_id = repo_id
        self.arch = arch
        self.platform_full_version = platform_full_version
        self.distributors_ids_type_ids_tuples = distributors_ids_type_ids


class Package(object):
    def __init__(self, name, filename, sourcerpm_filename=None, is_modular=False):
        self.name = name
        self.filename = filename
        self.sourcerpm_filename = sourcerpm_filename
        self.is_modular = is_modular

    def __lt__(self, other):
        return cmp_version(self.filename, other.filename) < 0

    def __gt__(self, other):
        return cmp_version(self.filename, other.filename) > 0

    def __eq__(self, other):
        return cmp_version(self.filename, other.filename) == 0

    def __le__(self, other):
        return cmp_version(self.filename, other.filename) <= 0

    def __ge__(self, other):
        return cmp_version(self.filename, other.filename) >= 0

    def __ne__(self, other):
        return cmp_version(self.filename, other.filename) != 0

    def __str__(self):
        return self.filename


class Module(object):
    def __init__(self, name, stream, version, context, arch, packages, profiles):
        self.name = name
        self.stream = stream
        self.version = version
        self.context = context
        self.arch = arch
        self.packages = packages
        self.profiles = profiles

    @property
    def nsvca(self):
        return ':'.join((self.name, self.stream, str(self.version), self.context, self.arch))

    def __str__(self):
        return self.nsvca


class ModuleDefaults(object):
    """
    module_defaults unit, defines which profiles are enabled by default when activating
    a module. For example:
    {...,
     "name": "ruby",
     "profiles": {
        "2.5": [
            "common"]
        },
    ...
    }
    if someone asks to enable 'ruby:2.5' for some repo without specifing profiles, will
    get 'common' profile by defualt
    """
    def __init__(self, name, stream, profiles):
        self.name = name
        self.stream = stream
        self.profiles = profiles  # a dict such as {'4.046':['common']}

    def __str__(self):
        return self.name

    @property
    def name_profiles(self):
        """
        flatten the profles and prepend name
        format: name:[key:profile,profile]:[key:profile]
        'ruby:[2.5:common,unique]'
        """
        result = self.name
        for key in sorted(self.profiles):
            result += ':[%s:%s]' % (key, ','.join(sorted(self.profiles[key])))
        return result




# borrowed from https://github.com/rpm-software-management/yum
def splitFilename(filename):
    """
    Pass in a standard style rpm fullname

    Return a name, version, release, epoch, arch, e.g.::
        foo-1.0-1.i386.rpm returns foo, 1.0, 1, i386
        1:bar-9-123a.ia64.rpm returns bar, 9, 123a, 1, ia64
    """

    if filename[-4:] == '.rpm':
        filename = filename[:-4]

    archIndex = filename.rfind('.')
    arch = filename[archIndex + 1:]

    relIndex = filename[:archIndex].rfind('-')
    rel = filename[relIndex + 1:archIndex]

    verIndex = filename[:relIndex].rfind('-')
    ver = filename[verIndex + 1:relIndex]

    epochIndex = filename.find(':')
    if epochIndex == -1:
        epoch = ''
    else:
        epoch = filename[:epochIndex]

    name = filename[epochIndex + 1:verIndex]
    return name, ver, rel, epoch, arch


class PulpAction(object):
    def __init__(self, units, dst_repo):
        self.units = units
        self.dst_repo = dst_repo

    def get_action(self, pulp_client_inst):
        raise NotImplementedError


class AssociateAction(PulpAction):
    def __init__(self, units, dst_repo, src_repo):
        super(AssociateAction, self).__init__(units, dst_repo)
        self.src_repo = src_repo

    def get_action(self, pulp_client_inst):
        raise NotImplementedError


class AssociateActionModules(AssociateAction):
    TYPE = "modules"

    def get_action(self, pulp_client_inst):
        return pulp_client_inst.associate_modules, self.src_repo, self.dst_repo, self.units


class UnassociateActionModules(PulpAction):
    TYPE = "modules"

    def get_action(self, pulp_client_inst):
        return pulp_client_inst.unassociate_modules, self.dst_repo, self.units


class AssociateActionModuleDefaults(AssociateAction):
    TYPE = "module_defaults"

    def get_action(self, pulp_client_inst):
        return pulp_client_inst.associate_module_defaults, self.src_repo, self.dst_repo, self.units


class UnassociateActionModuleDefaults(PulpAction):
    TYPE = "module_defaults"

    def get_action(self, pulp_client_inst):
        return pulp_client_inst.unassociate_module_defaults, self.dst_repo, self.units

class AssociateActionRpms(AssociateAction):
    TYPE = "packages"

    def get_action(self, pulp_client_inst):
        return pulp_client_inst.associate_packages, self.src_repo, self.dst_repo, self.units


class UnassociateActionRpms(PulpAction):
    TYPE = "packages"

    def get_action(self, pulp_client_inst):
        return pulp_client_inst.unassociate_packages, self.dst_repo, self.units



import re
import logging
import ubiconfig
from more_executors import Executors
from concurrent.futures import as_completed
from collections import namedtuple, defaultdict
from ubipop._pulp_client import Pulp, Package
from ubipop._utils import (splitFilename, AssociateActionModules,
                           AssociateActionModuleDefaults,
                           AssociateActionRpms,
                           UnassociateActionModules,
                           UnassociateActionModuleDefaults,
                           UnassociateActionRpms)
from itertools import chain

_LOG = logging.getLogger("ubipop")


class RepoMissing(Exception):
    pass


RepoSet = namedtuple('RepoSet', ['rpm', 'source', 'debug'])


class UbiRepoSet(object):
    def __init__(self, input_repos, output_repos):
        self.in_repos = input_repos
        self.out_repos = output_repos

        self.packages = defaultdict(list)
        self.debug_rpms = defaultdict(list)
        self.modules = defaultdict(list)
        self.module_defaults = defaultdict(list)
        self.pkgs_from_modules = defaultdict(list)
        self.source_rpms = defaultdict(list)

        self._ensure_repos_existence()

    def get_output_repo_ids(self):
        repos = set([self.out_repos.rpm.repo_id, self.out_repos.source.repo_id])
        if self.out_repos.debug:
            repos.add(self.out_repos.debug.repo_id)

        return repos

    def _ensure_repos_existence(self):
        fatal = False
        if not self.in_repos.rpm:
            fatal = True
            _LOG.error("Input Rpm repo does not exist")

        if not self.in_repos.source:
            fatal = True
            _LOG.error("Input Source repo does not exist")

        if not self.in_repos.debug:
            _LOG.warning("Input Debug repo does not exist")

        if not self.out_repos.rpm:
            fatal = True
            _LOG.error("Output Rpm repo does not exist")

        if not self.out_repos.source:
            fatal = True
            _LOG.error("Output Source repo does not exist")

        if not self.out_repos.debug:
            _LOG.warning("Output Debug repo does not exist")

        if fatal:
            raise RepoMissing()


class UbiPopulate(object):
    def __init__(self, pulp_hostname, pulp_auth, dry_run, ubiconfig_filename_list=None,
                 ubiconfig_dir_or_url=None, insecure=False, workers_count=4,
                 output_repos=None):

        self.ubiconfig_list = self._load_ubiconfig(ubiconfig_filename_list,
                                                   ubiconfig_dir_or_url)
        self.pulp = Pulp(pulp_hostname, pulp_auth, insecure)
        self.dry_run = dry_run
        self.output_repos = output_repos
        self._executor = Executors.thread_pool(max_workers=workers_count).with_retry()

    def _load_ubiconfig(self, filenames, ubiconfig_dir_or_url):
        loader = ubiconfig.get_loader(ubiconfig_dir_or_url)
        ubi_conf_list = []
        if filenames:
            ubi_conf_list.extend(loader.load(filename) for filename in filenames)
        else:
            ubi_conf_list.extend(loader.load_all())

        return ubi_conf_list

    def populate_ubi_repos(self):
        out_repos = set()

        for config in self.ubiconfig_list:
            try:
                repo_pairs = self._get_input_and_output_repo_pairs(config)
            except RepoMissing:
                _LOG.warning("Skipping current content triplet, some repos are missing")
                continue

            for repo_set in repo_pairs:
                UbiPopulateRunner(self.pulp, repo_set, config, self.dry_run,
                                  self._executor).run_ubi_population()

                out_repos.update(repo_set.get_output_repo_ids())

        if self.output_repos:
            with open(self.output_repos, 'w') as f:
                for repo in out_repos:
                    f.write(repo.strip() + '\n')

    def _get_input_and_output_repo_pairs(self, ubiconfig_item):
        """
        Determines pairs of input and output repos and also find correct source and debuginfo
        counterpart of repos.
        """

        rpms_cs = ubiconfig_item.content_sets.rpm
        source_cs = ubiconfig_item.content_sets.srpm
        debug_cs = ubiconfig_item.content_sets.debuginfo
        _LOG.info(
            "Getting input repos for input content sets:\n\t%s\n\t%s\n\t%s",
            rpms_cs.input, source_cs.input, debug_cs.input)

        in_repos_ft = self._executor.submit(self.pulp.search_repo_by_cs, rpms_cs.input)
        in_source_repos_ft = self._executor.submit(self.pulp.search_repo_by_cs, source_cs.input)
        in_debug_repos_ft = self._executor.submit(self.pulp.search_repo_by_cs, debug_cs.input)

        _LOG.info(
            "Getting output repos for output content sets:\n\t%s\n\t%s\n\t%s",
            rpms_cs.output, source_cs.output, debug_cs.output)

        out_repos_ft = self._executor.submit(self.pulp.search_repo_by_cs, rpms_cs.output)
        out_source_repos_ft = self._executor.submit(self.pulp.search_repo_by_cs, source_cs.output)
        out_debug_repos_ft = self._executor.submit(self.pulp.search_repo_by_cs, debug_cs.output)

        repo_pairs = []
        for input_repo in in_repos_ft.result():
            rpm = input_repo
            source = self._get_repo_counterpart(input_repo, in_source_repos_ft.result())
            debug_info = self._get_repo_counterpart(input_repo, in_debug_repos_ft.result())

            rhel_repo_set = RepoSet(rpm, source, debug_info)

            rpm = self._get_repo_counterpart(input_repo, out_repos_ft.result())
            source = self._get_repo_counterpart(input_repo, out_source_repos_ft.result())
            debug_info = self._get_repo_counterpart(input_repo, out_debug_repos_ft.result())

            ubi_repo_set = RepoSet(rpm, source, debug_info)

            repo_pairs.append(UbiRepoSet(rhel_repo_set, ubi_repo_set))

        return repo_pairs

    def _get_repo_counterpart(self, input_repo, repos_to_match):
        for repo in repos_to_match:
            if input_repo.arch == repo.arch and \
                    input_repo.platform_full_version == repo.platform_full_version:
                return repo


class UbiPopulateRunner(object):
    def __init__(self, pulp, output_repo_set, ubiconfig_item, dry_run, executor):
        self.pulp = pulp
        self.repos = output_repo_set
        self.ubiconfig = ubiconfig_item
        self.dry_run = dry_run
        self._executor = executor

    def _match_modules(self):
        # Add matching modules

        fts = {}
        for module in self.ubiconfig.modules:
            fts[self._executor.submit(self.pulp.search_modules,
                                      self.repos.in_repos.rpm, module.name,
                                      str(module.stream))] = \
                (module.name + str(module.stream), module.profiles)

        for ft in as_completed(fts):
            input_modules = ft.result()
            if input_modules:
                # fts[ft][0] == module.name + str(module.stream)
                # fts[ft][1] == module.profiles
                name_stream = fts[ft][0]
                profiles = fts[ft][1]
                self.repos.modules[name_stream].extend(input_modules)

                # Add packages from module profiles
                packages_names = self.get_packages_names_by_profiles(profiles, input_modules)

                for package_name in packages_names:
                    module_packages = self.get_packages_from_module(package_name, input_modules)

                    # for reference which pkgs are from modules
                    self.repos.pkgs_from_modules[name_stream].extend(module_packages)
                    self.repos.packages[package_name].extend(module_packages)

    def _match_module_defaults(self):
        """Try to find modulemd_defaults units in the same repo with the same
        name/stream of a modulemd.
        """
        fts = {}
        for _, modules in self.repos.modules.items():
            for md in modules:
                fts[self._executor.submit(self.pulp.search_module_defaults,
                                          self.repos.in_repos.rpm, md.name,
                                          str(md.stream))] = md.name + str(md.stream)
        for ft in as_completed(fts):
            module_defaults = ft.result()
            if module_defaults:
                self.repos.module_defaults[fts[ft]].extend(module_defaults)

    def _match_packages(self, repo, packages_dict):
        """
        Add matching packages from whitelist
        Globbing package name is not supported
        """
        fts = {}
        for package_pattern in self.ubiconfig.packages.whitelist:
            name = package_pattern.name
            arch = None if package_pattern.arch in ('*', None) else package_pattern.arch
            fts[(self._executor.submit(self.pulp.search_rpms,
                                       repo, name, arch))] = name

        for ft in as_completed(fts):
            packages = ft.result()
            if packages:
                packages_dict[fts[ft]].extend(packages)

    def _match_binary_rpms(self):
        self._match_packages(self.repos.in_repos.rpm, self.repos.packages)

    def _match_debug_rpms(self):
        self._match_packages(self.repos.in_repos.debug, self.repos.debug_rpms)

    def _parse_blacklist_config(self):
        packages_to_exclude = []
        for package_pattern in self.ubiconfig.packages.blacklist:
            name_to_parse = package_pattern.name
            globbing = '*' in name_to_parse
            if globbing:
                name = package_pattern.name[:-1]
            else:
                name = package_pattern.name
            arch = None if package_pattern.arch in ('*', None) else package_pattern.arch

            packages_to_exclude.append((name, globbing, arch))

        return packages_to_exclude

    def _exclude_blacklisted_packages(self):
        blacklisted_binary = self.get_blacklisted_packages(
            list(chain.from_iterable(self.repos.packages.values())))
        blacklisted_debug = self.get_blacklisted_packages(
            list(chain.from_iterable(self.repos.debug_rpms.values())))

        for pkg in blacklisted_binary:
            # blacklist only non-modular pkgs
            self.repos.packages[pkg.name][:] = [_pkg for _pkg in
                                                self.repos.packages.get(pkg.name, [])
                                                if _pkg.is_modular]

            # if there is nothing left, remove whole entry for package
            if not self.repos.packages[pkg.name]:
                self.repos.packages.pop(pkg.name, None)

        for pkg in blacklisted_debug:
            # blacklist only non-modular debug pkgs
            self.repos.debug_rpms[pkg.name][:] = [_pkg for _pkg in
                                                  self.repos.debug_rpms.get(pkg.name, [])
                                                  if _pkg.is_modular]

            # if there is nothing left, remove whole entry for debug package
            if not self.repos.debug_rpms[pkg.name]:
                self.repos.debug_rpms.pop(pkg.name, None)

    def _finalize_modules_output_set(self):
        for _, modules in self.repos.modules.items():
            self._finalize_output_units(modules, 'module')

    def _finalize_rpms_output_set(self):
        for _, packages in self.repos.packages.items():
            self._finalize_output_units(packages, 'rpm')

    def _finalize_debug_output_set(self):
        for _, packages in self.repos.debug_rpms.items():
            self._finalize_output_units(packages, 'rpm')

    def _finalize_output_units(self, units, type_id):
        if type_id == 'rpm':
            self.sort_packages(units)
            self.keep_n_latest_packages(units)  # with respect to packages referenced by modules
        else:
            self.sort_modules(units)
            self.keep_n_latest_modules(units)

    def _create_srpms_output_set(self):
        rpms = chain.from_iterable(self.repos.packages.values())
        for package in rpms:
            if package.sourcerpm_filename is None:
                _LOG.warning("Package %s doesn't reference its source rpm", package.name)
                continue

            self.repos.source_rpms[package.name].append(Package(package.name,
                                                                package.sourcerpm_filename))

        blacklisted_srpms = self.get_blacklisted_packages(
            list(chain.from_iterable(self.repos.source_rpms.values())))

        for pkg in blacklisted_srpms:
            self.repos.source_rpms.pop(pkg.name, None)

    def _determine_pulp_actions(self, units, current, diff_f):
        expected = list(chain.from_iterable(units.values()))
        to_associate = diff_f(expected, current)
        to_unassociate = diff_f(current, expected)
        return to_associate, to_unassociate

    def _get_pulp_actions_mds(self, modules, current):
        return self._determine_pulp_actions(modules, current, self._diff_modules_by_nsvca)

    def _get_pulp_actions_md_defaults(self, module_defaults, current):
        return self._determine_pulp_actions(module_defaults, current,
            self._diff_md_defaults_by_profiles)

    def _get_pulp_actions_pkgs(self, pkgs, current):
        return self._determine_pulp_actions(pkgs, current, self._diff_packages_by_filename)

    def _get_pulp_actions(self, current_modules_ft, current_module_defaults_ft, current_rpms_ft,
                          current_srpms_ft, current_debug_rpms_ft):
        """
        Determines expected pulp actions by comparing current content of output repos and
        expected content.

        Content that needs association: unit is in expected but not in current
        Content that needs unassociation: unit is in current but not in expected
        No action: unit is in current and in expected
        """
        modules_assoc, modules_unassoc = self._get_pulp_actions_mds(self.repos.modules,
                                                                    current_modules_ft.result())
        md_defaults_assoc, md_defaults_unassoc = \
            self._get_pulp_actions_md_defaults(self.repos.module_defaults,
                                               current_module_defaults_ft.result())

        rpms_assoc, rpms_unassoc = self._get_pulp_actions_pkgs(self.repos.packages,
                                                               current_rpms_ft.result())
        srpms_assoc, srpms_unassoc = self._get_pulp_actions_pkgs(self.repos.source_rpms,
                                                                 current_srpms_ft.result())

        debug_assoc = None
        debug_unassoc = None
        if current_debug_rpms_ft is not None:
            debug_assoc, debug_unassoc = self._get_pulp_actions_pkgs(self.repos.debug_rpms,
                                                                     current_debug_rpms_ft.result())

        associations = (AssociateActionModules(modules_assoc,
                                               self.repos.out_repos.rpm,
                                               self.repos.in_repos.rpm),
                        AssociateActionModuleDefaults(md_defaults_assoc,
                                                      self.repos.out_repos.rpm,
                                                      self.repos.in_repos.rpm),
                        AssociateActionRpms(rpms_assoc,
                                            self.repos.out_repos.rpm,
                                            self.repos.in_repos.rpm),
                        AssociateActionRpms(srpms_assoc,
                                            self.repos.out_repos.source,
                                            self.repos.in_repos.source),
                        AssociateActionRpms(debug_assoc,
                                            self.repos.out_repos.debug,
                                            self.repos.in_repos.debug)
                        )

        unassociations = (UnassociateActionModules(modules_unassoc, self.repos.out_repos.rpm),
                          UnassociateActionModuleDefaults(md_defaults_unassoc,
                                                          self.repos.out_repos.rpm),
                          UnassociateActionRpms(rpms_unassoc, self.repos.out_repos.rpm),
                          UnassociateActionRpms(srpms_unassoc, self.repos.out_repos.source),
                          UnassociateActionRpms(debug_unassoc, self.repos.out_repos.debug))

        return associations, unassociations

    def _diff_modules_by_nsvca(self, modules_1, modules_2):
        return self._diff_lists_by_attr(modules_1, modules_2, 'nsvca')

    def _diff_md_defaults_by_profiles(self, module_defaults_1, module_defaults_2):
        return self._diff_lists_by_attr(module_defaults_1, module_defaults_2, 'name_profiles')

    def _diff_packages_by_filename(self, packages_1, packages_2):
        return self._diff_lists_by_attr(packages_1, packages_2, 'filename')

    def _diff_lists_by_attr(self, list_1, list_2, attr):
        attrs_list_2 = [getattr(obj, attr) for obj in list_2]
        diff = [obj for obj in list_1 if getattr(obj, attr) not in attrs_list_2]

        return diff

    def run_ubi_population(self):
        current_modules_ft, current_module_defaults_ft, current_rpms_ft, \
            current_srpms_ft, current_debug_rpms_ft = self._get_current_content()

        self._match_modules()
        self._match_binary_rpms()
        if self.repos.out_repos.debug:
            self._match_debug_rpms()
        self._exclude_blacklisted_packages()
        self._finalize_modules_output_set()
        self._finalize_rpms_output_set()
        self._finalize_debug_output_set()
        self._match_module_defaults()
        self._create_srpms_output_set()

        associations, unassociations = self._get_pulp_actions(current_modules_ft,
                                                              current_module_defaults_ft,
                                                              current_rpms_ft,
                                                              current_srpms_ft,
                                                              current_debug_rpms_ft)

        if self.dry_run:
            self.log_curent_content(current_modules_ft, current_module_defaults_ft,
                                    current_rpms_ft, current_srpms_ft, current_debug_rpms_ft)
            self.log_pulp_actions(associations, unassociations)
        else:
            fts = []
            fts.extend(self._associate_unassociate_units(associations + unassociations))
            # wait for associate/unassociate tasks
            for ft in as_completed(fts):
                tasks = ft.result()
                if tasks:
                    self.pulp.wait_for_tasks(tasks)

            # wait repo publication
            for ft in as_completed(self._publish_out_repos()):
                self.pulp.wait_for_tasks(ft.result())

    def _associate_unassociate_units(self, action_list):
        fts = []
        for action in action_list:
            if action.units:
                fts.append(self._executor.submit(*action.get_action(self.pulp)))

        return fts

    def log_curent_content(self, current_modules_ft, current_module_defaults_ft,
                           current_rpms_ft, current_srpms_ft, current_debug_rpms_ft):
        _LOG.info("Current modules in repo: %s", self.repos.out_repos.rpm.repo_id)
        for module in current_modules_ft.result():
            _LOG.info(module.nsvca)

        _LOG.info("Current module_defaults in repo: %s", self.repos.out_repos.rpm.repo_id)
        for md_d in current_module_defaults_ft.result():
            _LOG.info("module_defaults: %s, profiles: %s", md_d.name, md_d.profiles)

        _LOG.info("Current rpms in repo: %s", self.repos.out_repos.rpm.repo_id)
        for rpm in current_rpms_ft.result():
            _LOG.info(rpm.filename)

        _LOG.info("Current srpms in repo: %s", self.repos.out_repos.source.repo_id)
        for rpm in current_srpms_ft.result():
            _LOG.info(rpm.filename)

        if self.repos.out_repos.debug:
            _LOG.info("Current rpms in repo: %s", self.repos.out_repos.debug.repo_id)
            for rpm in current_debug_rpms_ft.result():
                _LOG.info(rpm.filename)

    def log_pulp_actions(self, associations, unassociations):
        for item in associations:
            if item.units:
                for unit in item.units:
                    _LOG.info("Would associate %s from %s to %s", unit, item.src_repo.repo_id,
                              item.dst_repo.repo_id)

            else:
                _LOG.info("No association expected for %s from %s to %s", item.TYPE,
                          item.src_repo.repo_id,
                          item.dst_repo.repo_id)

        for item in unassociations:
            if item.units:
                for unit in item.units:
                    _LOG.info("Would unassociate %s from %s", unit, item.dst_repo.repo_id)
            else:
                _LOG.info("No unassociation expected for %s from %s", item.TYPE,
                          item.dst_repo.repo_id)

    def _get_current_content(self):
        """
        Gather current content of output repos
        """
        current_modules_ft = self._executor.submit(self.pulp.search_modules,
                                                   self.repos.out_repos.rpm)
        current_module_defaults_ft = self._executor.submit(self.pulp.search_module_defaults,
                                                           self.repos.out_repos.rpm)
        current_rpms_ft = self._executor.submit(self.pulp.search_rpms,
                                                self.repos.out_repos.rpm)
        current_srpms_ft = self._executor.submit(self.pulp.search_rpms,
                                                 self.repos.out_repos.source)
        if self.repos.out_repos.debug:
            current_debug_rpms_ft = self._executor.submit(self.pulp.search_rpms,
                                                          self.repos.out_repos.debug)
        else:
            current_debug_rpms_ft = None

        return current_modules_ft, current_module_defaults_ft, current_rpms_ft, \
            current_srpms_ft, current_debug_rpms_ft

    def _publish_out_repos(self):
        fts = []
        repos_to_publish = (self.repos.out_repos.rpm,
                            self.repos.out_repos.debug,
                            self.repos.out_repos.source)

        for repo in repos_to_publish:
            if repo:
                fts.append(self._executor.submit(self.pulp.publish_repo, repo))
        return fts

    def get_blacklisted_packages(self, package_list):
        """
        Finds blacklisted packages in output sets
        """
        blacklisted_pkgs = []
        for pattern_name, globbing, pattern_arch in self._parse_blacklist_config():
            for package in package_list:
                name, _, _, _, arch = splitFilename(package.filename)
                blacklisted = False
                if globbing:
                    if name.startswith(pattern_name):
                        blacklisted = True
                else:
                    if name == pattern_name:
                        blacklisted = True

                if pattern_arch:
                    if arch != pattern_arch:
                        blacklisted = False

                if blacklisted:
                    blacklisted_pkgs.append(package)

        return blacklisted_pkgs

    def sort_modules(self, modules):
        """
        Sort modules by version
        """
        modules.sort(key=lambda module: module.version)

    def keep_n_latest_modules(self, modules, n=1):
        """
        Keeps n latest modules in modules sorted list
        """
        modules[:] = modules[-n:]

    def sort_packages(self, packages):
        """
        Sort packages by vercmp
        """
        packages.sort()

    def get_packages_names_by_profiles(self, profiles, modules):
        """
        Gather package names by module profiles, if no profiles provided, add packages from
        all profiles

        Args:
            profiles (list of str):
                profiles ubi config
            modules (list of _pulp_client.Module):
                modules to process

        Returns:
            list of str:
                names of packages within matching modules & profiles
        """
        packages_names = []

        for module in modules:
            if profiles:
                for profile in profiles:
                    packages_names.extend(packages for packages in module.profiles.get(profile, []))
            else:
                for packages in module.profiles.values():
                    packages_names.extend(packages)

        return packages_names

    def get_packages_from_module(self, package_name, input_modules):
        """
        Gathers packages from modules by package name
        """
        rpms = []
        regex = r'\d+:'
        reg = re.compile(regex)
        for module in input_modules:
            for rpm_nevra in module.packages:
                rpm_without_epoch = reg.sub('', rpm_nevra)
                name, _, _, _, arch = splitFilename(rpm_without_epoch)
                # skip source package, they are calculated in later stage
                if arch == 'src':
                    continue
                if name == package_name:
                    rpms.append(Package(name, rpm_without_epoch + '.rpm', is_modular=True))

        return rpms

    def keep_n_latest_packages(self, packages, n=1):
        """
        Keep n latest non-modular packages,
        modular packages are kept only if they are referenced by some of remaining modules
        Parameter packages: sorted list of Packages objects, oldest goes first
        """
        packages_to_keep = []
        non_modular_pkgs = []
        for package in packages:
            if package.is_modular:
                for module_name_stream, packages_ref_by_module in \
                        self.repos.pkgs_from_modules.items():
                    pkgs_filenames_from_modules = [pkg.filename for pkg in packages_ref_by_module]
                    if package.filename in pkgs_filenames_from_modules \
                            and module_name_stream in self.repos.modules:
                        # this skips modular pkgs that are not referenced by module
                        packages_to_keep.append(package)
            else:
                non_modular_pkgs.append(package)

        # filter non-modular pkgs per arches, there can be rpms with different arches
        # for package in one repository
        pkgs_per_arch = defaultdict(list)
        for pkg in non_modular_pkgs:
            _, _, _, _, arch = splitFilename(pkg.filename)
            pkgs_per_arch[arch].append(pkg)

        latest_pkgs_per_arch = []
        for pkgs in pkgs_per_arch.values():
            latest_pkgs_per_arch += pkgs[-n:]

        packages[:] = latest_pkgs_per_arch + packages_to_keep



from setuptools import setup


def get_description():
    return 'Tool for populating ubi repositories'


def get_long_description():
    with open('README.md') as f:
        text = f.read()

    # Long description is everything after README's initial heading
    idx = text.find('\n\n')
    return text[idx:]


def get_requirements():
    with open('requirements.txt') as f:
        return f.read().splitlines()


setup(name='ubi-population-tool',
      version='0.1.13',
      license='GNU General Public License',
      author='',
      author_email='',
      description=get_description(),
      long_description=get_long_description(),
      long_description_content_type='text/markdown',
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.6',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Topic :: Software Development :: Libraries :: Python Modules',
      ],
      url="https://github.com/release-engineering/ubi-population-tool",
      install_requires=get_requirements(),
      packages=['ubipop'],
      entry_points={
          'console_scripts': [
              'ubipop = ubipop.cli:entry_point',
          ]
      }
      )
